//! Statistics functions.
pub mod kendall_tau;
pub use kendall_tau::weighted_kendall_tau_b;
pub mod min_max;
pub use min_max::max;
pub use min_max::min;
pub use min_max::min_max;
pub mod percentile;
pub use percentile::linear_percentile;
pub mod sample;
pub use sample::effective_sample_size;
pub mod sum;
pub use sum::sum;
pub mod sort;
pub use sort::weighted_merge_sort_mut;
